//
//  Tune.h
//  fab-video
//
//  Created by James Fletcher on 22/04/2014.
//
//

// Tune transposed from
// http://www.gamemusicthemes.com/sheetmusic/gameboy/tetris/themea/Tetris_-_Theme_A_by_Gori_Fater.pdf

#ifndef fab_video_Tune_h
#define fab_video_Tune_h

// Musical Notes
// Setup Timer0 count values to generate the frquencies for each note
// Note timer is 8 bit
#define _AUDIO_BASE_FREQ (F_CPU/1024)
#define REST 0
#if 0
#define A4 (_AUDIO_BASE_FREQ/440)
#define G4 (_AUDIO_BASE_FREQ/392)
#define F4 (_AUDIO_BASE_FREQ/349)
#define E4 (_AUDIO_BASE_FREQ/330)
#define D4 (_AUDIO_BASE_FREQ/294)
#define C4 (_AUDIO_BASE_FREQ/262) // Middle C approx 262Hz
#define B3 (_AUDIO_BASE_FREQ/247)
#define A3 (_AUDIO_BASE_FREQ/220)
#define G3 (_AUDIO_BASE_FREQ/196)
#define F3 (_AUDIO_BASE_FREQ/175)
#define E3 (_AUDIO_BASE_FREQ/165)
#define D3 (_AUDIO_BASE_FREQ/147)
#define C3 (_AUDIO_BASE_FREQ/131)
#define B2 (_AUDIO_BASE_FREQ/123)
#else
// Shift it all up an octave as I transposed it incorrectly
#define A4 (_AUDIO_BASE_FREQ/880)
#define G4 (_AUDIO_BASE_FREQ/784)
#define F4 (_AUDIO_BASE_FREQ/698)
#define E4 (_AUDIO_BASE_FREQ/659)
#define D4 (_AUDIO_BASE_FREQ/587)
#define C4 (_AUDIO_BASE_FREQ/523)
#define B3 (_AUDIO_BASE_FREQ/494)
#define A3 (_AUDIO_BASE_FREQ/440)
#define G3 (_AUDIO_BASE_FREQ/392)
#define F3 (_AUDIO_BASE_FREQ/349)
#define E3 (_AUDIO_BASE_FREQ/330)
#define D3 (_AUDIO_BASE_FREQ/294)
#define C3 (_AUDIO_BASE_FREQ/262)
#define B2 (_AUDIO_BASE_FREQ/247)
#define A2 (_AUDIO_BASE_FREQ/220)
#define G2 (_AUDIO_BASE_FREQ/196)
#define F2 (_AUDIO_BASE_FREQ/175)
#define E2 (_AUDIO_BASE_FREQ/165)
#define D2 (_AUDIO_BASE_FREQ/147)
#endif

// Note lengths as multiples of frame refresh rate
#define BREVE           120 //160
#define SEMIBREVE       60 //80
#define CROCHET         30 //40
#define QUAVER          15 //20
#define SEMIQUAVER      7 //10

// Tetris tune - a very poor approximation
typedef struct
{
    uint8_t pitch;
    uint8_t duration; // in video frames
} sNote;

// Tetris music
const sNote TUNE[] =
{
    {E4, CROCHET},
    {B3, QUAVER},
    {C4, QUAVER},
    {D4, QUAVER},
    {E4, SEMIQUAVER},
    {D4, SEMIQUAVER},
    {C4, QUAVER},
    {B3, QUAVER},
    
    {A3, CROCHET},
    {A3, QUAVER},
    {C4, QUAVER},
    {E4, CROCHET},
    {D4, QUAVER},
    {C4, QUAVER},
    
    {B3, (QUAVER*3)},
    {C4, QUAVER},
    {D4, CROCHET},
    {E4, CROCHET},
    
    {C4, CROCHET},
    {A3, CROCHET},
    {A3, SEMIBREVE},
    
    {D4, CROCHET},
    {F4, QUAVER},
    {A4, CROCHET},
    {G4, QUAVER},
    {F4, QUAVER},
    
    {E4, (QUAVER*3)},
    {C4, QUAVER},
    {E4, CROCHET},
    {D4, QUAVER},
    {C4, QUAVER},
    
    {B3, CROCHET},
    {B3, QUAVER},
    {C4, QUAVER},
    {D4, CROCHET},
    {E4, CROCHET},
    
    {C4, CROCHET},
    {A3, CROCHET},
    {A3, CROCHET},
    {REST, CROCHET},
    
    {E3, SEMIBREVE},
    {C3, SEMIBREVE},
    
    {D3, SEMIBREVE},
    {B2, SEMIBREVE},
    
    {C3, SEMIBREVE},
    {A2, SEMIBREVE},
    
    {G2, SEMIBREVE},
    {B2, SEMIBREVE},
    
    {E3, SEMIBREVE},
    {C3, SEMIBREVE},
    
    {D3, SEMIBREVE},
    {B2, SEMIBREVE},
    
    {C3, CROCHET},
    {E3, CROCHET},
    {A3, SEMIBREVE},
    
    {G3, BREVE},
    
    {REST, BREVE}
};

uint8_t     TUNE_length = sizeof(TUNE) / sizeof(sNote);
uint8_t     TUNE_pos;
uint32_t    TUNE_last_frame;


#endif
