/* Name: main.c
 * Author: <insert your name here>
 * Copyright: <insert your copyright message here>
 * License: <insert your license reference here>
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

//#include "Font.h"
//#include "Tune.h"

// Clock frequency defined in F_CPU from makefile
//#if F_CPU == 16000000
//#define _16MHz_ // set for Arduino rather than my Fab oard which runs at 20MHz
//#warning "16MHz device"
//#endif


// Video pins
#define VIDEO_port      PORTD
#define VIDEO_direction DDRD
#define VIDEO_low       _BV(PD2) // 1000R
#define VIDEO_high      _BV(PD1) // 499R
// Video Levels
#define VIDEO_SYNC      0
#define VIDEO_BLACK     (VIDEO_low)
#define VIDEO_GREY      (VIDEO_high)
#define VIDEO_WHITE     (VIDEO_high | VIDEO_low)

//
// Video timings (NTSC)
//
// Takes about 0.5us to process irq so add that to the relevant timings
// Can't generate less than 2.5us with the timer (40 instruction cycles)
// reserve registers and make irq context switch faster? Use low bits of actual state enum to directly write to some registers?
// In polling mode it takes around 1us to setup and process
#define _VIDEO_BASE_NS (1000000000/F_CPU) // cycle time of video base counter in ns
#define _VIDEO_setup_time           1000 // 1000 at 16MHz
#define _VIDEO_line_width           ((63500-(4*_VIDEO_setup_time))/_VIDEO_BASE_NS) // 63.5us full line width
#define _VIDEO_hsync_front_porch    ((1500-_VIDEO_setup_time)/_VIDEO_BASE_NS) // should really be 1.5us hsync front porch, but we can't time this
#define _VIDEO_hsync_width          ((4700-_VIDEO_setup_time)/_VIDEO_BASE_NS) // 4.7us hsync pulse width
#define _VIDEO_hsync_back_porch     ((4700-_VIDEO_setup_time)/_VIDEO_BASE_NS) // 4.7us hsync back porch
#define _VIDEO_line_pixel_tail      ((4700-_VIDEO_setup_time)/_VIDEO_BASE_NS) // 4.7us non-display pixel zone tail, muist be greater than tx time for one byte from SPI
#define _VIDEO_line_pixel_zone (_VIDEO_line_width - _VIDEO_hsync_front_porch - _VIDEO_hsync_width - _VIDEO_hsync_back_porch - _VIDEO_line_pixel_tail)
#define _VIDEO_baud                 2 // SPI baud rate register value used for video pixel clock, 1 gives > 160 pixels per line at fosc = 16MHz


// Audio output pin
// Line level drives at roughly 1v p-p, this will drive significantly lower with resistor network on PCB
#define AUDIO_port      PORTD
#define AUDIO_direction DDRD
#define AUDIO_pin       _BV(PD6)


#if 1
// Musical Notes
// Setup Timer0 count values to generate the frquencies for each note
// Note timer is 8 bit
#define _AUDIO_BASE_FREQ (F_CPU/1024)
#define REST 0
#if 0
#define A4 (_AUDIO_BASE_FREQ/440)
#define G4 (_AUDIO_BASE_FREQ/392)
#define F4 (_AUDIO_BASE_FREQ/349)
#define E4 (_AUDIO_BASE_FREQ/330)
#define D4 (_AUDIO_BASE_FREQ/294)
#define C4 (_AUDIO_BASE_FREQ/262) // Middle C approx 262Hz
#define B3 (_AUDIO_BASE_FREQ/247)
#define A3 (_AUDIO_BASE_FREQ/220)
#define G3 (_AUDIO_BASE_FREQ/196)
#define F3 (_AUDIO_BASE_FREQ/175)
#define E3 (_AUDIO_BASE_FREQ/165)
#define D3 (_AUDIO_BASE_FREQ/147)
#define C3 (_AUDIO_BASE_FREQ/131)
#define B2 (_AUDIO_BASE_FREQ/123)
#else
// Shift it all up an octave as I transposed it incorrectly
#define A4 (_AUDIO_BASE_FREQ/880)
#define G4 (_AUDIO_BASE_FREQ/784)
#define F4 (_AUDIO_BASE_FREQ/698)
#define E4 (_AUDIO_BASE_FREQ/659)
#define D4 (_AUDIO_BASE_FREQ/587)
#define C4 (_AUDIO_BASE_FREQ/523)
#define B3 (_AUDIO_BASE_FREQ/494)
#define A3 (_AUDIO_BASE_FREQ/440)
#define G3 (_AUDIO_BASE_FREQ/392)
#define F3 (_AUDIO_BASE_FREQ/349)
#define E3 (_AUDIO_BASE_FREQ/330)
#define D3 (_AUDIO_BASE_FREQ/294)
#define C3 (_AUDIO_BASE_FREQ/262)
#define B2 (_AUDIO_BASE_FREQ/247)
#define A2 (_AUDIO_BASE_FREQ/220)
#define G2 (_AUDIO_BASE_FREQ/196)
#define F2 (_AUDIO_BASE_FREQ/175)
#define E2 (_AUDIO_BASE_FREQ/165)
#define D2 (_AUDIO_BASE_FREQ/147)
#endif

// Note lengths as multiples of frame refresh rate
#define BREVE           120 //160
#define SEMIBREVE       60 //80
#define CROCHET         30 //40
#define QUAVER          15 //20
#define SEMIQUAVER      7 //10

// Tetris tune - a very poor approximation
typedef struct
{
    uint8_t pitch;
    uint8_t duration; // in video frames
} sNote;

// Tetris music
const sNote TUNE[] =
{
    {E4, CROCHET},
    {B3, QUAVER},
    {C4, QUAVER},
    {D4, QUAVER},
    {E4, SEMIQUAVER},
    {D4, SEMIQUAVER},
    {C4, QUAVER},
    {B3, QUAVER},
    
    {A3, CROCHET},
    {A3, QUAVER},
    {C4, QUAVER},
    {E4, CROCHET},
    {D4, QUAVER},
    {C4, QUAVER},
    
    {B3, (QUAVER*3)},
    {C4, QUAVER},
    {D4, CROCHET},
    {E4, CROCHET},
    
    {C4, CROCHET},
    {A3, CROCHET},
    {A3, SEMIBREVE},
    
    {D4, CROCHET},
    {F4, QUAVER},
    {A4, CROCHET},
    {G4, QUAVER},
    {F4, QUAVER},
    
    {E4, (QUAVER*3)},
    {C4, QUAVER},
    {E4, CROCHET},
    {D4, QUAVER},
    {C4, QUAVER},
    
    {B3, CROCHET},
    {B3, QUAVER},
    {C4, QUAVER},
    {D4, CROCHET},
    {E4, CROCHET},
    
    {C4, CROCHET},
    {A3, CROCHET},
    {A3, CROCHET},
    {REST, CROCHET},
    
    {E3, SEMIBREVE},
    {C3, SEMIBREVE},
    
    {D3, SEMIBREVE},
    {B2, SEMIBREVE},
    
    {C3, SEMIBREVE},
    {A2, SEMIBREVE},
    
    {G2, SEMIBREVE},
    {B2, SEMIBREVE},
    
    {E3, SEMIBREVE},
    {C3, SEMIBREVE},
    
    {D3, SEMIBREVE},
    {B2, SEMIBREVE},
    
    {C3, CROCHET},
    {E3, CROCHET},
    {A3, SEMIBREVE},
    
    {G3, BREVE},
    
    {REST, BREVE}
};

uint8_t     TUNE_length = sizeof(TUNE) / sizeof(sNote);
uint8_t     TUNE_pos;
uint32_t    TUNE_last_frame;
#endif



// Video global variables
static uint8_t VIDEO_port_temp;
static uint32_t VIDEO_frame_count;

// 144 x 80 pixel display
#define _VIDEO_display_width        144
#define _VIDEO_line_length_bytes    (_VIDEO_display_width/8) // Should force this to a power of 2 so we can speed up all our Drawing related multiplies
#define _VIDEO_display_height       80
static uint8_t VIDEO_buffer[_VIDEO_line_length_bytes*(_VIDEO_display_height+1)];

//
// Graphics drawing routines
// Caution, no range checking is provided on input coords!
//

// Set pixel
static inline void Draw_set_pixel(uint8_t x, uint8_t y)
{
    uint8_t index, mask;
    
    mask = x & 0x07;
    mask = 1<<mask;
    index = x>>3; // find the byte on the line
    VIDEO_buffer[(y*_VIDEO_line_length_bytes)+index] |= mask;
}

// Draw a line
// Based on Bresenham's
// Currently uses the set_pixel fn above, but could be greatly speed up with use of partial results
static void Draw_line(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
    int dx, dy;
    int sx, sy;
    int err;
    
    if(x0<x1)
    {
        dx = x1-x0;
        sx = 1;
    }
    else
    {
        dx = x0-x1;
        sx = -1;
    }
    
    if(y0<y1)
    {
        dy = y1-y0;
        sy = 1;
    }
    else
    {
        dy = y0-y1;
        sy = -1;
    }
    
    err = dx-dy;
    
    Draw_set_pixel(x0, y0);

    //while((x0!=x1) )// && (y0!=y1))
    while((y0!=y1))
    {
        if((err*2) > -dy)
        {
            err -= dy;
            x0 += sx;
        }
        if((err*2) < dx)
        {
            err += dx;
            y0 += sy;
        }
        Draw_set_pixel(x0, y0);
    }
}









//
// Processing Segments
//

// Process the tune during one of the vsync lines
static inline void Process_tune(void)
{
    // At vsync check for next note in tune
    // preset some of these vars to speed up tests?
    if(VIDEO_frame_count > (TUNE_last_frame + TUNE[TUNE_pos].duration))
    {
        TUNE_pos++;
        if(TUNE_pos >= TUNE_length)
            TUNE_pos = 0;
        TUNE_last_frame = VIDEO_frame_count;
        if(TUNE[TUNE_pos].pitch == 0)
            TCCR0B = 0b00000000; // clock stopped
        else
        {
            TCCR0B = 0b00000101; // 1024 prescaler
            OCR0A = TUNE[TUNE_pos].pitch;
        }
    }
}

// Clear frame buffer
// Caution this will almost certainly take too long in one blast
static inline void Process_clear_frame(uint8_t start_line, uint8_t end_line)
{
    uint8_t i,j;
    
    //uint32_t *p = (uint32_t*)(&(VIDEO_buffer[start_line*_VIDEO_line_length_bytes])); // can't do this as line starts aren't u32 aligned!
    uint8_t *p = &(VIDEO_buffer[start_line*_VIDEO_line_length_bytes]);
    
    for(i=start_line; i<end_line; i++)
    {
        for(j=0; j<_VIDEO_line_length_bytes; j+=4)
        {
            *p++ = 0;
            *p++ = 0;
            *p++ = 0;
            *p++ = 0;
        }
    }
}

// Animate something
int tx[3];
int ty[3];
int dtx[3];
int dty[3];
static void Process_animate(void)
{
    static uint8_t x = 0, y = 0;
    uint8_t i;
    
    Draw_set_pixel(x, y);
    x++;
    if(x >= _VIDEO_display_width) x=0;
    y++;
    if(y >= _VIDEO_display_height) y=0;
    
    //Draw_line(10,10, 30,50);
    
    for(i=0; i<3; i++)
    {
        tx[i] += dtx[i];
        if(tx[i] < 0) { tx[i] = 0; dtx[i] = -dtx[i]; }
        if(tx[i] >= _VIDEO_display_width) { tx[i] = _VIDEO_display_width; dtx[i] = -dtx[i]; }

        ty[i] += dty[i];
        if(ty[i] < 0) { ty[i] = 0; dty[i] = -dty[i]; }
        if(ty[i] >= _VIDEO_display_height) { ty[i] = _VIDEO_display_height; dty[i] = -dty[i]; }
    }
}

static void Process_render1(void)
{
    Draw_line(tx[0], ty[0], tx[1], ty[1]);
    //Draw_line(tx[1], ty[1], tx[2], ty[2]);
    //Draw_line(tx[2], ty[2], tx[0], ty[0]);
}

static void Process_render2(void)
{
    //Draw_line(tx[0], ty[0], tx[1], ty[1]);
    Draw_line(tx[1], ty[1], tx[2], ty[2]);
    //Draw_line(tx[2], ty[2], tx[0], ty[0]);
}

static void Process_render3(void)
{
    //Draw_line(tx[0], ty[0], tx[1], ty[1]);
    //Draw_line(tx[1], ty[1], tx[2], ty[2]);
    Draw_line(tx[2], ty[2], tx[0], ty[0]);
}


//
// Video line and frame segments
//

// Front porch 1.5us - set black: 0.3v
static inline void front_porch(uint8_t level)
{
    VIDEO_port_temp = VIDEO_port & ~(VIDEO_high | VIDEO_low);
    VIDEO_port = VIDEO_port_temp | level; //VIDEO_low;
    TCNT1H = (0xffff-_VIDEO_hsync_front_porch) >> 8;
    TCNT1L = (0xffff-_VIDEO_hsync_front_porch) & 0xff;
    TIFR1 |= _BV(TOV1);
    while((TIFR1 & _BV(TOV1)) == 0)
        ;
}

// Hsync - set sync: 0v
static inline void hsync(uint8_t level)
{
    VIDEO_port_temp = VIDEO_port & ~(VIDEO_high | VIDEO_low);
    VIDEO_port = VIDEO_port_temp | level;
    TCNT1H = (0xffff-_VIDEO_hsync_width) >> 8;
    TCNT1L = (0xffff-_VIDEO_hsync_width) & 0xff;
    TIFR1 |= _BV(TOV1);
    while((TIFR1 & _BV(TOV1)) == 0)
        ;
}

// Back porch - set black: 0.3v
static inline void back_porch(uint8_t level)
{
    VIDEO_port_temp = VIDEO_port & ~(VIDEO_high | VIDEO_low);
    VIDEO_port = VIDEO_port_temp | level; //VIDEO_low;
    TCNT1H = (0xffff-_VIDEO_hsync_back_porch) >> 8;
    TCNT1L = (0xffff-_VIDEO_hsync_back_porch) & 0xff;
    TIFR1 |= _BV(TOV1); // clear overflow flag
    while((TIFR1 & _BV(TOV1)) == 0) // spin waiting for count to complete
        ;
}

// A blank pixel zone, used in sync lines - set black: 0.3v
static inline void pixel_blank(uint8_t level, uint8_t processing_segment)
{
    VIDEO_port_temp = VIDEO_port & ~(VIDEO_high | VIDEO_low);
    VIDEO_port = VIDEO_port_temp | level; //VIDEO_low;
    TCNT1H = (0xffff-_VIDEO_line_pixel_zone) >> 8;
    TCNT1L = (0xffff-_VIDEO_line_pixel_zone) & 0xff;
    TIFR1 |= _BV(TOV1); // clear overflow flag
    
    // Do some processing.  Each prcessing segment must take less than 50us
    switch(processing_segment)
    {
            //_VIDEO_display_height
        case 0:
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:
        case 7:
            Process_clear_frame(processing_segment*10, ((processing_segment+1)*10)-1);
            break;
        case 10:
            Process_tune();
            break;
        case 11:
            Process_animate();
            break;
        case 12:
            Process_render1();
            break;
        case 13:
            Process_render2();
            break;
        case 14:
            Process_render3();
            break;
        default:
            break;
    }
    
    while((TIFR1 & _BV(TOV1)) == 0) // spin waiting for count to complete
        ;
}

// Non-display pixel tail - set black: 0.3v - If you set to anything other than black after a real pixel zone then you must wait for SPI tx to finish first
static inline void pixel_tail(uint8_t level)
{
    VIDEO_port_temp = VIDEO_port & ~(VIDEO_high | VIDEO_low);
    VIDEO_port = VIDEO_port_temp | level; //VIDEO_low;
    TCNT1H = (0xffff-_VIDEO_line_pixel_tail) >> 8;
    TCNT1L = (0xffff-_VIDEO_line_pixel_tail) & 0xff;
    TIFR1 |= _BV(TOV1); // clear overflow flag
    while((TIFR1 & _BV(TOV1)) == 0) // spin waiting for count to complete
        ;
}


int main(void)
{
    uint8_t i;
    uint8_t count_temp;
    uint8_t line_byte_index;
    uint8_t processing_segment;
    
    // Initialise Timer0 to produce audio tones on OCA0
    // counts at sysclk/1024 (15.625KHz at 16MHz, 19.531KHz at 20MHz)
    // Set CTC mode with OC0A toggling on each match
    TCCR0A = 0b01000010; // CTC mode
    //TCCR0B = 0b00000101; // 1024 prescaler
    TCCR0B = 0b00000000; // clock stopped
    TCNT0 = 0;
    OCR0A = C4;
    // Set Audio ouput - OC0A
    AUDIO_direction |= AUDIO_pin;
    
    // Set Video output pins
    VIDEO_direction |= (VIDEO_low | VIDEO_high);
    // Set both low
    VIDEO_port &= (~(VIDEO_low | VIDEO_high));

    // Initialise Timer1 to produce video timings
    // Counts up in Normal mode
    // No OC1x behaviour
    TCCR1A = 0b00000000;
    TCCR1B = 0b00000001; // Normal, 1 prescaler
    TCNT1H = (0xffff-_VIDEO_line_width) >> 8;
    TCNT1L = (0xffff-_VIDEO_line_width) & 0xff;
    //TIMSK1 |= _BV(TOIE1); // Enable timer 1 overflow irq
    
    // Initialise USART0 in SPI mode
    UBRR0 = 0;
    /* Setting the XCKn port pin as output, enables master mode. */
    DDRD |= (1<<PD4);
    /* Set MSPI mode of operation and SPI data mode 0. */
    UCSR0C = (1<<UMSEL01)|(1<<UMSEL00)|(1<<UDORD0)|(0<<UCPHA0)|(0<<UCPOL0);
    /* Enable receiver and transmitter. */
    // UCSRnB = (1<<RXEN0)|(1<<TXEN0); not yet
    /* Set baud rate. */
    /* IMPORTANT: The Baud Rate must be set after the transmitter is enabled */
    //UBRR0 = 0; // Baud as high as possible, roughly 0.125us per bit
    UBRR0 = _VIDEO_baud;
        
    // Initialise frame buffer with a simple pattern
    //for(i=0; i<_VIDEO_display_height; i++)
    //{
    //    VIDEO_buffer[i*(_VIDEO_line_length_bytes+1)] = 0xaa;
    //}
    
    // Initialise test triangle
    tx[0] = 10; ty[0] = 10;
    tx[1] = 30; ty[1] = 50;
    tx[2] = 80; ty[2] = 30;
    
    dtx[0] = 2; dty[0] = 1;
    dtx[1] = 1; dty[1] = 1;
    dtx[2] = -1; dty[2] = -1;
    
    VIDEO_frame_count = 0;
    TUNE_pos = 0;
    TUNE_last_frame = 0;
    
    for(;;)
    {
        processing_segment = 0;
        
        // Generate 3 blank black lines
        for(i=0; i<3; i++)
        {
            front_porch(VIDEO_BLACK);
            hsync(VIDEO_SYNC);
            back_porch(VIDEO_BLACK);
            pixel_blank(VIDEO_BLACK, processing_segment);
            pixel_tail(VIDEO_BLACK);
            processing_segment++;
        }

        // Generate 3 VSync lines
        for(i=0; i<3; i++)
        {
            front_porch(VIDEO_SYNC);
            hsync(VIDEO_WHITE);
            back_porch(VIDEO_SYNC);
            pixel_blank(VIDEO_SYNC, processing_segment);
            pixel_tail(VIDEO_SYNC);
            processing_segment++;
        }
        
        //
        // We actually increase the black lines at the start of frame and reduce the displayed lines.
        // Without this the lines at the start of frame don't get displayed on all devices.
        // Also gives me more processing segments.
        //
        
        // Generate 12 blank black lines
        for(i=0; i<20/*12*/; i++)
        {
            front_porch(VIDEO_BLACK);
            hsync(VIDEO_SYNC);
            back_porch(VIDEO_BLACK);
            pixel_blank(VIDEO_BLACK, processing_segment);
            pixel_tail(VIDEO_BLACK);
            processing_segment++;
        }

        // Generate 242 standard display lines
        for(i=0; i<234/*242*/; i++) // i is actually the index of the current display line
        {
            front_porch(VIDEO_BLACK);
            hsync(VIDEO_SYNC);
            back_porch(VIDEO_BLACK);
        
            // Video pixel zone
            VIDEO_port_temp = VIDEO_port & ~(VIDEO_high | VIDEO_low);
            VIDEO_port = VIDEO_port_temp | VIDEO_GREY; // set grey: 0.6v
            TCNT1H = (0xffff-_VIDEO_line_pixel_zone) >> 8;
            TCNT1L = (0xffff-_VIDEO_line_pixel_zone) & 0xff;
            TIFR1 |= _BV(TOV1); // clear overflow flag
        
            UCSR0B = (1<<TXEN0); // enable USART tx
            UBRR0 = _VIDEO_baud; // Restart baud rate generator so that pixels at start of line are aligned
       //     UDR0 = 0xaa;
            UDR0 = 0x00;
            VIDEO_port = VIDEO_port_temp | VIDEO_WHITE; // set white: 1.0v
        
       //     count_temp = 0;
            line_byte_index = 0;
            while((TIFR1 & _BV(TOV1)) == 0) // spin waiting for count to complete
            {
                uint8_t byte = 0;
                
                if((i>>1) < _VIDEO_display_height)
                {
                    if(line_byte_index < _VIDEO_line_length_bytes)
                    {
                        byte = VIDEO_buffer[((i>>1)*_VIDEO_line_length_bytes) + line_byte_index];
                        line_byte_index++;
                    }
                }
                
       //         if(count_temp & 0x01)
       //             UDR0 = 0xaa;
       //         else
       //             UDR0 = 0x00;
       //         count_temp++;

                UDR0 = byte;
                
                while(((TIFR1 & _BV(TOV1)) == 0) && ((UCSR0A & _BV(UDRE0)) == 0))// spin waiting for count to complete or SPI tx data reg empty
                    ;
            }
        
            UCSR0B = 0; // disable USART tx
        
            pixel_tail(VIDEO_BLACK);
        }

        VIDEO_frame_count++;
    }
    return 0;   /* never reached */
}
